# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from operator import or_
from typing import Tuple, Dict, Optional, NamedTuple

from sqlalchemy import func, select
from sqlalchemy.orm import sessionmaker

from i2fhirb2.fhir.fhirspecific import DEFAULT_PROJECT_ID, IDE_SOURCE_HIVE, DEFAULT_PATIENT_NUMBER_START
from i2fhirb2.i2b2model.data.i2b2patientmapping import PatientMapping, PatientIDEStatus
from i2fhirb2.sqlsupport.dbconnection import I2B2Tables


class PatientNumberGenerator:
    """
    i2b2 patient number generator.
    """
    def __init__(self, next_number: int) -> None:
        self._next_number = next_number

    def new_number(self) -> int:
        rval = self._next_number
        self._next_number += 1
        return rval

    def refresh(self, tables: I2B2Tables, ignore_upload_id: Optional[int]) -> int:
        session = sessionmaker(bind=tables.crc_engine)()
        q = func.max(tables.patient_dimension.c.patient_num)
        if ignore_upload_id is not None:
            q = q.filter(or_(tables.patient_dimension.c.upload_id.is_(None),
                         tables.patient_dimension.c.upload_id != ignore_upload_id))
        qr = session.query(q).all()
        self._next_number = qr[0][0] + 1
        session.close()
        return self._next_number


class FHIRPatientMapping:
    project_id = DEFAULT_PROJECT_ID         # Default project identifier
    identity_source_id = IDE_SOURCE_HIVE    # source_id for identity mapping
    number_generator = PatientNumberGenerator(DEFAULT_PATIENT_NUMBER_START)
    number_map = dict()                     # type: Dict[Tuple[str, str, str], int]

    class PatientMappingKey(NamedTuple):
        patient_id: str
        patient_ide_src: str
        project_id: str

    @classmethod
    def _clear(cls):
        cls.project_id = DEFAULT_PROJECT_ID
        cls.identity_source_id = IDE_SOURCE_HIVE
        cls.number_generator = PatientNumberGenerator(DEFAULT_PATIENT_NUMBER_START)
        cls.number_map.clear()

    @classmethod
    def refresh_patient_number_generator(cls, tables: I2B2Tables, ignore_upload_id: Optional[int]) -> int:
        return cls.number_generator.refresh(tables, ignore_upload_id)

    def __init__(self, tables: Optional[I2B2Tables], patient_id: str, patient_ide_source: str) -> None:
        """
        Create a new patient mapping entry in the FHIR context
        :param tables: i2b2 data tables
        :param patient_id: identifier
        :param patient_ide_source: source -- currently the base URI
        """
        self.patient_mapping_entries = []
        key = self.PatientMappingKey(patient_id, patient_ide_source, self.project_id)
        if key in self.number_map:
            self.patient_num = self.number_map[key]
        else:
            self.patient_num = self._existing_entry(tables, key)
            if self.patient_num is None:
                self.patient_num = self.number_generator.new_number()
            pm = PatientMapping(self.patient_num,
                                patient_id,
                                PatientIDEStatus.active,
                                patient_ide_source,
                                self.project_id)
            self.number_map[key] = self.patient_num
            self.patient_mapping_entries.append(pm)

        identity_id = str(self.patient_num)
        ikey = self.PatientMappingKey(identity_id, self.identity_source_id, self.project_id)
        if ikey not in self.number_map:
            ipm = PatientMapping(self.patient_num,
                                 identity_id,
                                 PatientIDEStatus.active,
                                 self.identity_source_id,
                                 self.project_id)
            self.number_map[ikey] = ipm
            self.patient_mapping_entries.append(ipm)

    @staticmethod
    def _existing_entry(tables: Optional[I2B2Tables], key: "FHIRPatientMapping.PatientMappingKey") -> Optional[int]:
        if not tables:
            return None                                 # Generating tsv files, nothihc exists
        pdtabc = tables.patient_mapping.c
        s = select([pdtabc.patient_num])\
            .where(pdtabc.patient_ide == key.patient_id and
                   pdtabc.patient_ide_source == key.patient_ide_src
                   and pdtabc.project_id == key.project_id)
        qr = list(tables.crc_connection.execute(s))
        return qr[0][0] if qr else None



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Optional, cast, Sized


class VisualAttributes:
    """ i2b2 Metadata Visual attributes element
    """
    def __init__(self, string_value: Optional[str] = None) -> None:
        """
        Constructor from optional string value.
        :param string_value:
        """
        assert (string_value is None or len(string_value) in (2, 3))
        if string_value is not None:
            char1 = string_value[0]
            assert (char1 in "CFMLODR")
            self.leaf = char1 in 'LMR'
            self.approximate = char1 == 'M'
            self.draggable = char1 in 'FD'
            self.concept = char1 in 'CFML'

            char2 = string_value[1]
            assert (char2 in "AIH")
            self.active = char2 == 'A'
            self.hidden = char2 == 'H'

            char3 = string_value[2] if len(cast(Sized, string_value)) == 3 else ' '
            assert (char3 in " E")
            self.editable = char3 == 'E'
        else:
            self.leaf = False           # C
            self.approximate = False    # Leaf only - multiple targets
            self.draggable = True       # Folder only
            self.concept = True         # Concept or Modifier Dimension

            self.active = True
            self.hidden = False         # No commitment whether hidden is active

            self.editable = True

    def __str__(self):
        if self.concept:
            if self.leaf:
                rval = 'M' if self.approximate else 'L'
            else:
                rval = 'F' if self.draggable else 'C'
        else:
            if self.leaf:
                rval = 'R'
            else:
                rval = 'D' if self.draggable else 'O'
        rval += 'H' if self.hidden else 'A' if self.active else 'I'
        rval += 'E' if self.editable else ' '
        return rval



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from argparse import Namespace
from datetime import datetime
from typing import List, Tuple, Optional

import sys
from fhirtordf.rdfsupport.fhirgraphutils import value
from fhirtordf.rdfsupport.uriutils import uri_to_ide_and_source
from rdflib import Graph, RDF, URIRef

from i2fhirb2.fhir.fhirobservationfact import FHIRObservationFactFactory
from i2fhirb2.fhir.fhirpatientdimension import FHIRPatientDimension
from i2fhirb2.fhir.fhirpatientmapping import FHIRPatientMapping
from i2fhirb2.fhir.fhirresourcemap import FHIR_RESOURCE_MAP, FHIR_Infrastructure_type, FHIR_Observation_Fact_type, \
    FHIR_Visit_Dimension_type, FHIR_Provider_Dimension_type, FHIR_Patient_Dimension_type, FHIR_Bundle_type, \
    FHIR_Resource_type
from i2fhirb2.fhir.fhirspecific import FHIR
from i2fhirb2.fhir.fhirvisitdimension import FHIRVisitDimension
from i2fhirb2.i2b2model.data.i2b2encountermapping import EncounterMapping
from i2fhirb2.i2b2model.data.i2b2observationfact import ObservationFactKey, ObservationFact
from i2fhirb2.i2b2model.data.i2b2patientdimension import PatientDimension
from i2fhirb2.i2b2model.data.i2b2patientmapping import PatientMapping
from i2fhirb2.i2b2model.data.i2b2visitdimension import VisitDimension
from i2fhirb2.i2b2model.metadata.i2b2conceptdimension import ConceptDimension
from i2fhirb2.i2b2model.metadata.i2b2modifierdimension import ModifierDimension
from i2fhirb2.i2b2model.shared.i2b2core import I2B2_Core
from i2fhirb2.sqlsupport.dbconnection import I2B2Tables
from i2fhirb2.sqlsupport.i2b2tables import change_column_length
from i2fhirb2.tsv_support.tsvwriter import write_tsv


# TODO: Handle continuation pages in queries and bundles

class I2B2GraphMap:
    def __init__(self, g: Graph, opts: Namespace) -> None:
        """
        Iterate over the resources in the graph mapping them to their i2b2 equivalent
        :param g: graph
        :param opts: input options
        """
        self._opts = opts
        self._g = g
        self.num_infrastructure = 0         # Number of infrastructure resources encountered (and not loaded)
        self.num_visit = 0                  # Number of visit resources (to be implemented)
        self.num_provider = 0               # Number of provider resources
        self.num_bundle = 0                 # Number of bundler resources (we should be unwrapping these?)
        self.num_unmapped = 0               # Number of untyped resources encountered (need classifying)
        self.observation_facts = []         # type: List[ObservationFact]
        self.patient_dimensions = []        # type: List[PatientDimension]
        self.patient_mappings = []          # type: List[PatientMapping]
        self.visit_dimensions = []          # type: List[VisitDimension]
        self.encounter_mappings = []        # type: List[EncounterMapping]
        self.tables = opts.tables           # type: I2B2Tables
        nresources = 0
        for subj, subj_type in g.subject_objects(RDF.type):
            if isinstance(subj, URIRef) and subj_type in FHIR_RESOURCE_MAP:
                print("{}: ({}) - {}".format(nresources, str(subj_type).split('/')[-1], subj))
                nresources += 1
                mapped_type = FHIR_RESOURCE_MAP[subj_type]
                if isinstance(mapped_type, FHIR_Infrastructure_type):
                    self.num_infrastructure += 1
                elif isinstance(mapped_type, FHIR_Observation_Fact_type):
                    pm, vd, start_date = self.process_resource_instance(subj, mapped_type)
                    if pm is not None:
                        obsfactory = \
                            FHIRObservationFactFactory(g, ObservationFactKey(pm.patient_num,
                                                                             vd.visit_dimension_entry.encounter_num,
                                                                             opts.providerid, start_date), subj)
                        # TODO: Decide what do do with the other mappings in the observation factory
                        self.observation_facts += obsfactory.observation_facts
                elif isinstance(mapped_type, FHIR_Visit_Dimension_type):
                    self.num_visit += 1
                elif isinstance(mapped_type, FHIR_Provider_Dimension_type):
                    self.num_provider += 1
                elif isinstance(mapped_type, FHIR_Patient_Dimension_type):
                    pd = FHIRPatientDimension(self._g, self.tables, subj)
                    self.patient_dimensions.append(pd.patient_dimension_entry)
                    self.patient_mappings += pd.patient_mappings.patient_mapping_entries
                elif isinstance(mapped_type, FHIR_Bundle_type):
                    self.num_bundle += 1
                else:
                    self.num_unmapped += 1
        print("---> Graph map phase complete")

    def process_resource_instance(self, subj: URIRef,  mapped_type: FHIR_Resource_type) \
            -> Tuple[Optional[FHIRPatientMapping], Optional[FHIRVisitDimension], Optional[datetime]]:
        patient_id_uri, encounter_id_uri, provider_id = mapped_type.fact_key_for(self._g, subj)
        if patient_id_uri is not None:
            patient_id, patient_ide_source = uri_to_ide_and_source(patient_id_uri)
            pm = FHIRPatientMapping(self.tables, patient_id, patient_ide_source)
            self.patient_mappings += pm.patient_mapping_entries
            start_date = value(self._g, subj, FHIR.Observation.effectiveDateTime)
            if not start_date:
                start_date = datetime.now()
            vd = FHIRVisitDimension(subj, pm.patient_num, patient_id, patient_ide_source, start_date)
            self.visit_dimensions.append(vd.visit_dimension_entry)
            self.encounter_mappings += vd.encounter_mappings.encounter_mapping_entries
            return pm, vd, start_date
        else:
            # Just a reference to a resource instance -- drop it
            return None, None, None

    def generate_tsv_files(self) -> None:
        self._generate_tsv_file("observation_fact.tsv", ObservationFact, self.observation_facts)
        self._generate_tsv_file("patient_dimension.tsv", PatientDimension, self.patient_dimensions)
        self._generate_tsv_file("patient_mapping.tsv", PatientMapping, self.patient_mappings)
        self._generate_tsv_file("visit_dimension.tsv", VisitDimension, self.visit_dimensions)
        self._generate_tsv_file("encounter_mapping.tsv", EncounterMapping, self.encounter_mappings)

    def _generate_tsv_file(self, fname: str, cls, values: List[I2B2_Core]) -> None:
        write_tsv(self._opts.outdir, fname, cls._header(), values)

    @staticmethod
    def clear_i2b2_tables(tables: I2B2Tables, uploadid: int) -> None:
        """
        Remove all entries in the i2b2 tables for uploadid.
        :param tables:
        :param uploadid:
        :return:
        """
        # This is a static function to support the removefacts operation
        print("Deleted {} patient_dimension records"
              .format(PatientDimension.delete_upload_id(tables, uploadid)))
        print("Deleted {} patient_mapping records"
              .format(PatientMapping.delete_upload_id(tables, uploadid)))
        print("Deleted {} observation_fact records"
              .format(ObservationFact.delete_upload_id(tables, uploadid)))
        print("Deleted {} visit_dimension records"
              .format(VisitDimension.delete_upload_id(tables, uploadid)))
        print("Deleted {} encounter_mapping records"
              .format(EncounterMapping.delete_upload_id(tables, uploadid)))

    @staticmethod
    def clear_i2b2_sourcesystems(tables: I2B2Tables, sourcesystemcd: str) -> None:
        print("Deleted {} patient_dimension records"
              .format(PatientDimension.delete_sourcesystem_cd(tables, sourcesystemcd)))
        print("Deleted {} patient_mapping records"
              .format(PatientMapping.delete_sourcesystem_cd(tables, sourcesystemcd)))
        print("Deleted {} observation_fact records"
              .format(ObservationFact.delete_sourcesystem_cd(tables, sourcesystemcd)))
        print("Deleted {} visit_dimension records"
              .format(VisitDimension.delete_sourcesystem_cd(tables, sourcesystemcd)))
        print("Deleted {} encounter_mapping records"
              .format(EncounterMapping.delete_sourcesystem_cd(tables, sourcesystemcd)))

    def load_i2b2_tables(self, check_dups=False) -> None:
        # session = sessionmaker(bind=tables.crc_engine)()
        I2B2_Core._check_dups = check_dups
        if self._opts.remove:
            # TODO: This should really be within a transaction boundary
            self.clear_i2b2_tables(self._opts.tables, self._opts.uploadid)
        change_column_length(self._opts.tables.observation_fact, self._opts.tables.observation_fact.c.concept_cd,
                             200, self._opts.tables.crc_connection)
        change_column_length(self._opts.tables.observation_fact, self._opts.tables.observation_fact.c.modifier_cd,
                             200, self._opts.tables.crc_connection)
        print("{} / {} patient_dimension records added / modified"
              .format(*PatientDimension.add_or_update_records(self._opts.tables, self.patient_dimensions)))
        print("{} / {} patient_mapping records added / modified"
              .format(*PatientMapping.add_or_update_records(self._opts.tables, self.patient_mappings)))
        print("{} / {} visit_dimension records added / modified"
              .format(*VisitDimension.add_or_update_records(self._opts.tables, self.visit_dimensions)))
        print("{} / {} encounter_mapping records added / modified"
              .format(*EncounterMapping.add_or_update_records(self._opts.tables, self.encounter_mappings)))
        print("{} / {} observation_fact records added / modified"
              .format(*ObservationFact.add_or_update_records(self._opts.tables, self.observation_facts)))
        # session.close()

    def summary(self) -> str:
        summary_text = """Generated:
    {} Observation facts
    {} Patients
    {} Patient mappings
"""
        skip_text = """=== SKIPS ===
    {num_bundle} Bundled resources (shouldn't happen?)
    {num_visit} Visit resources
    {num_infrastructure} Infrastructure resources
    {num_provider} Provider resources
    {num_unmapped} Unmapped resources
"""
        num_skips = self.num_infrastructure + self.num_visit + self.num_provider + self.num_unmapped + self.num_bundle
        rval = summary_text.format(len(self.observation_facts),
                                   len(self.patient_dimensions),
                                   len(self.patient_mappings))
        if num_skips:
            rval += skip_text.format(**self.__dict__)
        return rval



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.


# create table i2b2demodata.patient_mapping
# (
# 	patient_ide varchar(200) not null,
# 	patient_ide_source varchar(50) not null,
# 	patient_num integer not null,
# 	patient_ide_status varchar(50),
# 	project_id varchar(50) not null,
# 	upload_date timestamp,
# 	update_date timestamp,
# 	download_date timestamp,
# 	import_date timestamp,
# 	sourcesystem_cd varchar(50),
# 	upload_id integer,
# 	constraint patient_mapping_pk
# 		primary key (patient_ide, patient_ide_source, project_id)
# )
# ;
from typing import List, Tuple


from i2fhirb2.i2b2model.shared.i2b2core import I2B2_Core_With_Upload_Id
from i2fhirb2.sqlsupport.dynobject import DynElements, DynObject
from i2fhirb2.sqlsupport.dbconnection import I2B2Tables


class PatientIDEStatus:
    class PatientIDEStatusCode:
        def __init__(self, code: str) -> None:
            self.code = code
    active = PatientIDEStatusCode("A")
    inactive = PatientIDEStatusCode("I")
    deleted = PatientIDEStatusCode("D")
    merged = PatientIDEStatusCode("M")


class PatientMapping(I2B2_Core_With_Upload_Id):
    _t = DynElements(I2B2_Core_With_Upload_Id)

    key_fields = ["patient_ide", "patient_ide_source", "project_id"]

    def __init__(self, patient_num: int, patient_id: str, patient_ide_status: PatientIDEStatus.PatientIDEStatusCode,
                 patient_ide_source: str, project_id: str, **kwargs):
        """
        Construct a patient mapping entry
        :param patient_num: patient number
        :param patient_id: clear text patient identifier
        :param patient_ide_status: status code
        :param patient_ide_source: clear text patient identifier source
        :param project_id: project identifier
        :param kwargs:

        The patient number is the key to the patient dimension file.  The patient_mapping file has, at a minimum,
        one entry
        """
        self._patient_num = patient_num
        self._patient_ide = patient_id
        self._patient_ide_status = patient_ide_status
        self._patient_ide_source = patient_ide_source
        self._project_id = project_id

        super().__init__(**kwargs)

    @DynObject.entry(_t)
    def patient_ide(self) -> str:
        """
        The encrypted patient identifier
        """
        return self._patient_ide

    @DynObject.entry(_t)
    def patient_ide_source(self) -> str:
        """
        The source system (source of what I don't really know)
        """
        return self._patient_ide_source

    @DynObject.entry(_t)
    def patient_num(self) -> int:
        """
        Patient number in the patient dimension file
        """
        return self._patient_num

    @DynObject.entry(_t)
    def patient_ide_status(self) -> str:
        """
        Patient number in the patient dimension file
        """
        return self._patient_ide_status.code

    @DynObject.entry(_t)
    def project_id(self) -> str:
        """
        Project identifier
        """
        return self._project_id

    @classmethod
    def delete_upload_id(cls, tables: I2B2Tables, upload_id: int) -> int:
        """
        Delete all patient_mapping records with the supplied upload_id
        :param tables: i2b2 sql connection
        :param upload_id: upload identifier to remove
        :return: number or records that were deleted
        """
        return cls._delete_upload_id(tables.crc_connection, tables.patient_mapping, upload_id)

    @classmethod
    def delete_sourcesystem_cd(cls, tables: I2B2Tables, sourcesystem_cd: str) -> int:
        """
        Delete all records with the supplied sourcesystem_cd
        :param tables: i2b2 sql connection
        :param sourcesystem_cd: sourcesystem_cd to remove
        :return: number or records that were deleted
        """
        return cls._delete_sourcesystem_cd(tables.crc_connection, tables.patient_mapping, sourcesystem_cd)

    @classmethod
    def add_or_update_records(cls, tables: I2B2Tables, records: List["PatientMapping"]) -> Tuple[int, int]:
        """
        Add or update the patient_mapping table as needed to reflect the contents of records
        :param tables: i2b2 sql connection
        :param records: records to apply
        :return: number of records added / modified
        """
        return cls._add_or_update_records(tables.crc_connection, tables.patient_mapping, records)



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from fhirtordf.rdfsupport.namespaces import namespace_for
from rdflib import URIRef
from rdflib.namespace import split_uri

from i2fhirb2.fhir.fhirspecific import concept_path, concept_code, concept_name
from i2fhirb2.i2b2model.shared.i2b2core import I2B2_Core_With_Upload_Id
from i2fhirb2.sqlsupport.dynobject import DynElements


class CommonDimension(I2B2_Core_With_Upload_Id):
    """ Common base class of all dimensions """
    _t = DynElements(I2B2_Core_With_Upload_Id)
    graph = None

    def __init__(self, subject: URIRef, base_path: str = '\\', **kwargs) -> None:
        """
        Constructor
        :param subject: URI of the subject
        :param base_path: base path of items
        :param kwargs: additional arguments for i2b2 core
        """
        super().__init__(**kwargs)
        assert(base_path.endswith('\\'))
        self._subject = subject
        self._base_path = base_path
        self._name_prefix = namespace_for(split_uri(subject)[0]).upper()

    def path(self) -> str:
        return self._base_path + concept_path(self._subject)

    def cd(self) -> str:
        return concept_code(self._subject)

    def name_char_(self) -> str:
        return self._name_prefix + ' ' + concept_name(self.graph, self._subject)

    @staticmethod
    def blob() -> str:
        return ''



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from datetime import datetime
from typing import Optional, Tuple, List


from i2fhirb2.i2b2model.shared.i2b2core import I2B2_Core_With_Upload_Id
from i2fhirb2.sqlsupport.dynobject import DynElements, DynObject
from i2fhirb2.sqlsupport.dbconnection import I2B2Tables


class ObservationFactKey:
    def __init__(self, patient_num: int, encounter_num: int, provider_id: str, start_date: Optional[datetime] = None) -> None:
        """
        A partial key to the observation fact table.  These three or four elements identify a collection of facts
        that can be added or replaced as a block
        :param patient_num: Patient number -- key to the patient_dimension table
        :param encounter_num: Encounter number -- key to the visit_dimension table
        :param provider_id: Provider identifier -- key to the provider_dimension table
        :param start_date: If provided, separates multiple entries w/ different dates of occurrence
        """
        self.patient_num = patient_num
        self.encounter_num = encounter_num
        self.provider_id = provider_id
        self.key_includes_start_date = start_date is not None
        self.start_date = start_date if start_date is not None else datetime.now()


class ValueTypeCd:
    def __init__(self, code: str) -> None:
        self.code = code


valuetype_text = ValueTypeCd('T')
valuetype_number = ValueTypeCd('N')
valuetype_blob = ValueTypeCd('B')
valuetype_nlp = ValueTypeCd('NLP')
valuetype_date = ValueTypeCd('D')
valuetype_novalue = ValueTypeCd('@')


# create table i2b2demodata.observation_fact
# (
#     encounter_num integer not null,
#     patient_num integer not null,
#     concept_cd varchar(50) not null,
#     provider_id varchar(50) not null,
#     start_date timestamp not null,
#     modifier_cd varchar(100) default '@'::character varying not null,
#     instance_num integer default 1 not null,
#     valtype_cd varchar(50),
#     tval_char varchar(255),
#     nval_num numeric(18,5),
#     valueflag_cd varchar(50),
#     quantity_num numeric(18,5),
#     units_cd varchar(50),
#     end_date timestamp,
#     location_cd varchar(50),
#     observation_blob text,
#     confidence_num numeric(18,5),
#     update_date timestamp,
#     download_date timestamp,
#     import_date timestamp,
#     sourcesystem_cd varchar(50),
#     upload_id integer,
#     text_search_index serial not null,
#     constraint observation_fact_pk
#         primary key (patient_num, concept_cd, modifier_cd, start_date, encounter_num, instance_num, provider_id)
# )

class ObservationFact(I2B2_Core_With_Upload_Id):
    _t = DynElements(I2B2_Core_With_Upload_Id)

    key_fields = ["patient_num", "concept_cd", "modifier_cd", "start_date",
                  "encounter_num", "instance_num", "provider_id"]

    def __init__(self, fact_key: ObservationFactKey, concept_cd: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self._patient_num = fact_key.patient_num
        self._encounter_num = fact_key.encounter_num
        self._provider_id = fact_key.provider_id
        self._concept_cd = concept_cd
        self._start_date = fact_key.start_date
        self._modifier_cd = None
        self._instance_num = None
        self._valtype_cd = None
        self._tval_char = None
        self._nval_num = None
        self._valueflag_cd = None
        self._quantity_num = None
        self._units_cd = None
        self._end_date = None
        self._location_cd = None
        self._observation_blob = None
        self._confidence_num = None

    @DynObject.entry(_t)
    def encounter_num(self) -> int:
        """
        Encoded i2b2 patient visit number
        """
        return self._encounter_num

    @DynObject.entry(_t)
    def patient_num(self) -> int:
        """
        Encoded i2b2 patient number
        """
        return self._patient_num

    @DynObject.entry(_t)
    def concept_cd(self) -> str:
        """ Code for the observation of interest (i.e. diagnoses, procedures, medications, lab tests"""
        return self._concept_cd

    @DynObject.entry(_t)
    def provider_id(self) -> str:
        """
        Practitioner or provider id
        :return:
        """
        return self._provider_id

    @DynObject.entry(_t)
    def start_date(self) -> datetime:
        """ Starting date-time of the observation. (mm/dd/yy)"""
        return self._start_date

    @DynObject.entry(_t)
    def modifier_cd(self) -> str:
        """
        Code for modifier of interest (i.e. "Route", "DOSE").
        Note that the value columns are often used to hold the amounts such as "100" (mg) for the modifier of DOSE or
        "PO" for the modifier of ROUTE.
        :return:
        """
        return self._modifier_cd if self._modifier_cd else '@'

    @DynObject.entry(_t)
    def instance_num(self) -> int:
        """
        Encoded instance number that allows more than one modifier to be provided for each CONCEPT_CD.
        Each row will have a different MODIFIER_CD but a similar INSTANCE_NUM
        """
        return self._instance_num if self._instance_num is not None else 0

    @DynObject.entry(_t)
    def valtype_cd(self) -> str:
        """ The place where the value is stored
         'T' -- tval_char (Text (enums / short messages)
         'N' -- nval_num has number, tval_char has optional G, E, L (or other stuff where needed)
         'B' -- Raw text (notes / reports) observation_blob
         'NLP' -- NLP result text
         '@' -- no value
         'D' -- tval_char has a date in text and nval has a floating point representation (
                1996-10-13 00:00 = 19961013.0000
         ''  -- pure nval_num (?) - no tval_char?
         Note that the spec says that this is optional, but it is never null
          """
        return self._valtype_cd if self._valtype_cd is not None else '@'

    @DynObject.entry(_t)
    def tval_char(self) -> Optional[str]:
        """
        Used in conjunction with VALTYPE_CD = "T" or "N"
        When "T"
            Stores the text value
        When "N"
            'E' - Equals, 'NE' - Not equal, 'L' - Less than, 'LE' - LE, 'G' - Greater than, 'GE
        :return:
        """
        return self._tval_char

    @DynObject.entry(_t)
    def nval_num(self) -> Optional[float]:
        """
        Used in conjunction with VALTYPE_CD = "N" to store a numeric value
        """
        return self._nval_num

    @DynObject.entry(_t)
    def valueflag_cd(self) -> Optional[str]:
        """
        A code that represents the type of value present in the NVAL_NUM, the TVAL_CHAR or OBSERVATION_BLOB columns
        'X' - Encrypted text in the blob column (B)
        'H' - High      (N or T)
        'L' - Low
        'A' - Abnormal
        """
        return self._valueflag_cd

    @DynObject.entry(_t)
    def quantity_num(self) -> Optional[float]:
        """
        The number of observations represented by this fact (no known examples)
        """
        return self._quantity_num

    @DynObject.entry(_t)
    def units_cd(self) -> Optional[str]:
        """
        A textual description of the units associated with a value
        :return:
        """
        return self._units_cd

    @DynObject.entry(_t)
    def end_date(self) -> Optional[datetime]:
        """
        The date that the observation ended. If the date is derived or calculated from another observation (like a
        report) then the end date is the same as the observation it was derived or calculated from
        WARNING: two meanings for same field
        """
        return self._end_date

    @DynObject.entry(_t)
    def location_cd(self) -> Optional[str]:
        """
        A code representing the hospital associated with this visit
        """
        return self._location_cd

    @DynObject.entry(_t)
    def observation_blob(self) -> Optional[str]:
        """
        XML data that includes partially structured and unstructured data about an observation
        """
        return self._observation_blob

    @DynObject.entry(_t)
    def confidence_num(self) -> Optional[float]:
        """
        A code or number representing the confidence in the accuracy of the data. (No known examples)
        Q: 'code' in a numeric situation?
        """
        return None

    DynObject._after_root(_t)

    # text_search_index is auto-generated (?)
    # @DynObject.entry(_t)
    # def text_search_index(self) -> float:
    #     return None
    @property
    def pk(self) -> Tuple:
        return self.patient_num, self.encounter_num, self.instance_num,  self.concept_cd, self.modifier_cd

    def __lt__(self, other: "ObservationFact") -> bool:
        return self.pk < other.pk

    @classmethod
    def delete_upload_id(cls, tables: I2B2Tables, upload_id: int) -> int:
        """
        Delete all observation_fact records with the supplied upload_id
        :param tables: i2b2 sql connection
        :param upload_id: upload identifier to remove
        :return: number or records that were deleted
        """
        return cls._delete_upload_id(tables.crc_connection, tables.observation_fact, upload_id)

    @classmethod
    def delete_sourcesystem_cd(cls, tables: I2B2Tables, sourcesystem_cd: str) -> int:
        """
        Delete all records with the supplied sourcesystem_cd
        :param tables: i2b2 sql connection
        :param sourcesystem_cd: sourcesystem_cd to remove
        :return: number or records that were deleted
        """
        return cls._delete_sourcesystem_cd(tables.crc_connection, tables.observation_fact, sourcesystem_cd)

    @classmethod
    def add_or_update_records(cls, tables: I2B2Tables, records: List["ObservationFact"]) -> Tuple[int, int]:
        """
        Add or update the observation_fact table as needed to reflect the contents of records
        :param tables: i2b2 sql connection
        :param records: records to apply
        :return: number of records added / modified
        """
        return cls._add_or_update_records(tables.crc_connection, tables.observation_fact, records)

    def _date_val(self, dt: datetime) -> None:
        """
        Add a date value
        :param dt: datetime to add
        """
        self._tval_char = dt.strftime('%Y-%m-%d %H:%M')
        self._nval_num = (dt.year * 10000) + (dt.month * 100) + dt.day + \
                         (((dt.hour / 100.0) + (dt.minute / 10000.0)) if isinstance(dt, datetime) else 0)



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
import argparse
import shlex
from argparse import Namespace
from typing import List

import os

from i2fhirb2.fhir.fhirspecific import DEFAULT_ONTOLOGY_TABLE
from i2fhirb2.i2b2model.shared.tablenames import i2b2tablenames
from i2fhirb2.sqlsupport.i2b2tables import I2B2Tables

Default_DB_Connection = "postgresql+psycopg2://localhost:5432/i2b2"
Default_User = "postgres"
Default_Password = "postgres"


class FileAwareParser(argparse.ArgumentParser):
    def __init__(self, use_defaults: bool=True, *args, **kwargs) -> None:
        self.use_defaults = use_defaults
        super().__init__(*args, **kwargs)
        self.file_args = []

    def add_file_argument(self, *args, **kwargs):
        rval = self.add_argument(*args, **kwargs)
        self.file_args.append(rval)
        return rval

    def add_argument(self, *args, **kwargs):
        defhelp = kwargs.pop("help", None)
        defaults = kwargs.pop("default", None)
        default = defaults if self.use_defaults else None
        if not defhelp or default is None or kwargs.get('action') == 'help':
            return super().add_argument(*args, help=defhelp, default=default, **kwargs)
        else:
            return super().add_argument(*args, help=defhelp + " (default: {})".format(default),
                                        default=default, **kwargs)


class ConfigFile(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs) -> None:
        super().__init__(option_strings, dest, nargs, **kwargs)

    def __call__(self, parser: FileAwareParser, namespace: Namespace, values, option_string=None):
        raise AttributeError("Must preprocess input arguments with decode_file_args function")


def add_connection_args(parser: FileAwareParser) -> FileAwareParser:
    """
    Add the database connection arguments to the supplied parser

    :param parser: parser to add arguments to
    :return: parser
    """
    parser.add_file_argument("--conf", metavar="CONFIG FILE", help="Configuration file", action=ConfigFile)
    parser.add_argument("-db", "--dburl", help="Default database URL",
                        default=Default_DB_Connection)
    parser.add_argument("--user", help="Default user name",
                        default=Default_User)
    parser.add_argument("--password", help="Default password",
                        default=Default_Password)
    parser.add_argument("--crcdb", help="CRC database URL. (default: dburl)")
    parser.add_argument("--crcuser", help="User name for CRC database. (default: user)")
    parser.add_argument("--crcpassword", help="Password for CRC database. (default: password)")
    parser.add_argument("--ontodb", help="Ontology database URL.  (default: dburl)")
    parser.add_argument("--ontouser", help="User name for ontology database. (default: user)")
    parser.add_argument("--ontopassword", help="Password for ontology database. (default: password)")
    parser.add_argument("--onttable", metavar="ONTOLOGY TABLE NAME",
                        help="Ontology table name (default: {})".format(DEFAULT_ONTOLOGY_TABLE),
                        default=DEFAULT_ONTOLOGY_TABLE)
    return parser


def process_parsed_args(opts: Namespace, connect: bool=True) -> Namespace:
    """
    Set the defaults for the crc and ontology schemas
    :param opts: parsed arguments
    :param connect: actually connect. (For debugging)
    :return: namespace with additional elements added
    """
    def setdefault(vn: str, default: object) -> None:
        assert vn in opts, "Unknown option"
        if not getattr(opts, vn):
            setattr(opts, vn, default)

    setdefault('crcdb', opts.dburl)
    setdefault('crcuser', opts.user)
    setdefault('crcpassword', opts.password)
    setdefault('ontodb', opts.dburl)
    setdefault('ontouser', opts.user)
    setdefault('ontopassword', opts.password)
    if connect:
        opts.tables = I2B2Tables(opts)

    i2b2tablenames.ontology_table = opts.onttable
    return opts


def decode_file_args(argv: List[str], parser: FileAwareParser = None) -> List[str]:
    """
    Preprocess a configuration file.  The location of the configuration file is stored in the parser so that the
    FileOrURI action can add relative locations.
    :param argv: raw options list
    :param parser: argument parser.  Used as an anchor for the configuration file location
    :return: options list with '--conf' references replaced with file contents
    """
    for i in range(0, len(argv) - 1):
        # TODO: take prefix into account
        if argv[i] == '--conf':
            del argv[i]
            conf_file = argv[i]
            del(argv[i])
            with open(conf_file) as config_file:
                conf_args = shlex.split(config_file.read())
                if parser:
                    argv += fix_rel_paths(conf_args, parser, conf_file)
            return decode_file_args(argv)
    return argv


def fix_rel_paths(conf_args: List[str], parser: FileAwareParser, conf_file: str) -> List[str]:
    base_path = os.path.abspath(os.path.split(conf_file)[0])
    rval = []
    is_file_arg = False
    for conf_arg in conf_args:
        if any(conf_arg.startswith(prefix) for prefix in parser.prefix_chars):
            rval.append(conf_arg)
            is_file_arg = any(conf_arg in file_action.option_strings for file_action in parser.file_args)
            # TODO: recursive config files
        elif is_file_arg and ('://' not in conf_arg and not os.path.isabs(conf_arg)):
                rval.append(os.path.abspath(os.path.join(base_path, conf_arg)))
        else:
            rval.append(conf_arg)
    return rval



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

import collections


# Taken from https://stackoverflow.com/questions/1653970/does-python-have-an-ordered-set#1653974
class OrderedSet(collections.OrderedDict, collections.MutableSet):
    """
    An Ordered Set - combines the uniqueness of a set with entry ordering
    """
    def update(self, *args, **kwargs):
        if kwargs:
            raise TypeError("update() takes no keyword arguments")

        for s in args:
            for e in s:
                self.add(e)

    def add(self, elem):
        self[elem] = None

    def discard(self, elem):
        self.pop(elem, None)

    def __le__(self, other):
        return all(e in other for e in self)

    def __lt__(self, other):
        return self <= other and self != other

    def __ge__(self, other):
        return all(e in self for e in other)

    def __gt__(self, other):
        return self >= other and self != other

    def __repr__(self):
        return 'OrderedSet([%s])' % (', '.join(map(repr, self.keys())))

    def __str__(self):
        return '{%s}' % (', '.join(map(repr, self.keys())))

    difference = property(lambda self: self.__sub__)
    difference_update = property(lambda self: self.__isub__)
    intersection = property(lambda self: self.__and__)
    intersection_update = property(lambda self: self.__iand__)
    issubset = property(lambda self: self.__le__)
    issuperset = property(lambda self: self.__ge__)
    symmetric_difference = property(lambda self: self.__xor__)
    symmetric_difference_update = property(lambda self: self.__ixor__)
    union = property(lambda self: self.__or__)



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from argparse import Namespace
from typing import Tuple, List

from sqlalchemy import MetaData, create_engine, Table, Column
from sqlalchemy.engine import Engine

from i2fhirb2.i2b2model.shared.tablenames import i2b2tablenames


class I2B2Tables:
    """
    Class that represents the set of i2b2 tables that are used in this package.  The actual tables
    can be accessed by:
        t = I2B2Tables(opts)
        t.visit_dimension       # Specific table
        t.crc_engine            # Engine for crc tables
        t.crc_connection        # Connection for crc tables
    """
    i2b2metadata = 'i2b2metadata'
    i2b2crc = 'i2b2demodata'

    def __init__(self, opts: Namespace) -> None:
        _metadata = MetaData()
        crc_url, ont_url = self._db_urls(opts)

        self.crc_engine = create_engine(crc_url)
        self.crc_connection = self.crc_engine.connect()
        _metadata.reflect(bind=self.crc_engine, schema=self.i2b2crc)
        self._crc_tables = _metadata.tables
        if ont_url != crc_url:
            self.ont_engine = create_engine(ont_url)
            self.ont_connection = self.ont_engine.connect()
        else:
            self.ont_engine = self.crc_engine
            self.ont_connection = self.crc_connection

        _metadata.reflect(bind=self.ont_engine, schema=self.i2b2metadata)
        self._ont_tables = _metadata.tables

    # Note: If you get a recursion error below, you've got an unitialized self variable in the __init__ section
    def __getattr__(self, item):
        phys_name = i2b2tablenames.phys_name(item)
        k = I2B2Tables.i2b2metadata + '.' + phys_name
        if k in self._ont_tables:
            return self._ont_tables[k]
        k = I2B2Tables.i2b2crc + '.' + phys_name
        if k in self._crc_tables:
            return self._crc_tables[k]
        return None

    def __getitem__(self, item):
        return getattr(self, item)

    @staticmethod
    def _db_urls(opts: Namespace) -> Tuple[str, str]:
        """
        Return the crc and ontology db urls
        :param opts: options
        :return: Tuple w/ crc and ontology url
        """
        return opts.crcdb.replace("//", "//{crcuser}:{crcpassword}@".format(**opts.__dict__)),\
            opts.ontodb.replace("//", "//{ontouser}:{ontopassword}@".format(**opts.__dict__))

    def _tables(self) -> List[Tuple[str, str]]:
        """
        Return a list of all known tables and and its full URI
        :return: table name and full URI
        """
        return [(k.rsplit('.', 1)[1] if '.' in k else k, k) for k in self._ont_tables.keys()]


def change_column_length(table: Table, column: Column, length: int, engine: Engine) -> None:
    """ Change the column length in the supplied table
    """
    if column.type.length < length:
        print("Changing length of {} from {} to {}".format(column, column.type.length, length))
        column.type.length = length
        column_name = column.name
        column_type = column.type.compile(engine.dialect)
        engine.execute('ALTER TABLE {table} ALTER COLUMN {column_name} TYPE {column_type}'.format(**locals()))



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from datetime import datetime
from typing import Optional, List, Tuple

from i2fhirb2.i2b2model.shared.i2b2core import I2B2_Core_With_Upload_Id
from i2fhirb2.sqlsupport.dynobject import DynElements, DynObject
from i2fhirb2.sqlsupport.dbconnection import I2B2Tables


class ActiveStatusCd:
    def __setattr__(self, key, value):
        if key not in ["startcode", "endcode"]:
            raise ValueError("New elements not allowed")
        super().__setattr__(key, value)

    class EndDateCode:
        def __init__(self, code: str) -> None:
            self.code = code

    ed_unknown = EndDateCode('U')
    ed_ongoing = EndDateCode('O')
    ed_day = EndDateCode('Y')
    ed_month = EndDateCode('M')
    ed_year = EndDateCode('X')
    ed_hour = EndDateCode('R')
    ed_minute = EndDateCode('T')
    ed_second = EndDateCode('S')

    class StartDateCode:
        def __init__(self, code: str) -> None:
            self.code = code

    sd_unknown = StartDateCode('L')
    sd_ongoing = StartDateCode('A')
    sd_day = StartDateCode('D')
    sd_month = StartDateCode('B')
    sd_year = StartDateCode('F')
    sd_hour = StartDateCode('H')
    sd_minute = StartDateCode('I')
    sd_second = StartDateCode('C')

    def __init__(self, start: StartDateCode, end: EndDateCode) -> None:
        self.startcode = start
        self.endcode = end

    @property
    def code(self):
        return self.endcode.code + self.startcode.code


class VisitDimension(I2B2_Core_With_Upload_Id):
    _t = DynElements(I2B2_Core_With_Upload_Id)

    key_fields = ["encounter_num", "patient_num"]

    def __init__(self,
                 encounter_num: int,
                 patient_num: int,
                 active_status_cd: Optional[ActiveStatusCd] = None,
                 start_date: Optional[datetime] = None,
                 end_date: Optional[datetime] = None,
                 inout_cd: Optional[str] = None,
                 location_cd: Optional[str] = None,
                 location_path: Optional[str] = None,
                 length_of_stay: Optional[int] = None,
                 visit_blob: Optional[str] = None,
                 **kwargs):
        self._encounter_num = encounter_num
        self._patient_num = patient_num
        self._active_status_cd = active_status_cd
        self._start_date = start_date
        self._end_date = end_date
        self._inout_cd = inout_cd
        self._location_cd = location_cd
        self._location_path = location_path
        self._length_of_stay = length_of_stay
        self._visit_blob = visit_blob
        super().__init__(**kwargs)

    @DynObject.entry(_t)
    def encounter_num(self) -> int:
        """
        Reference number for the encounter/visit
        """
        return self._encounter_num

    @DynObject.entry(_t)
    def patient_num(self) -> int:
        """
        Reference number for the patient
        """
        return self._patient_num

    @DynObject.entry(_t)
    def active_status_cd(self) -> Optional[str]:
        """
        Reference number for the patient
        """
        return self._active_status_cd.code if self._active_status_cd else None

    @DynObject.entry(_t)
    def start_date(self) -> Optional[datetime]:
        """
        The date the event began
        """
        return self._start_date

    @DynObject.entry(_t)
    def end_date(self) -> Optional[datetime]:
        """
        The date the event ended
        """
        return self._end_date

    @DynObject.entry(_t)
    def inout_cd(self) -> Optional[str]:
        """
        The date the event ended
        """
        return self._inout_cd

    @DynObject.entry(_t)
    def location_cd(self) -> Optional[str]:
        return self._location_cd

    @DynObject.entry(_t)
    def location_path(self) -> Optional[str]:
        return self._location_path

    @DynObject.entry(_t)
    def length_of_stay(self) -> Optional[int]:
        return self._length_of_stay

    @DynObject.entry(_t)
    def visit_blob(self) -> Optional[str]:
        return self._visit_blob

    @classmethod
    def delete_upload_id(cls, tables: I2B2Tables, upload_id: int) -> int:
        """
        Delete all patient_dimension records with the supplied upload_id
        :param tables: i2b2 sql connection
        :param upload_id: upload identifier to remove
        :return: number or records that were deleted
        """
        return cls._delete_upload_id(tables.crc_connection, tables.visit_dimension, upload_id)

    @classmethod
    def delete_sourcesystem_cd(cls, tables: I2B2Tables, sourcesystem_cd: str) -> int:
        """
        Delete all records with the supplied sourcesystem_cd
        :param tables: i2b2 sql connection
        :param sourcesystem_cd: sourcesystem_cd to remove
        :return: number or records that were deleted
        """
        return cls._delete_sourcesystem_cd(tables.crc_connection, tables.visit_dimension, sourcesystem_cd)

    @classmethod
    def add_or_update_records(cls, tables: I2B2Tables, records: List["VisitDimension"]) -> Tuple[int, int]:
        """
        Add or update the patient_dimension table as needed to reflect the contents of records
        :param tables: i2b2 sql connection
        :param records: records to apply
        :return: number of records added / modified
        """
        return cls._add_or_update_records(tables.crc_connection, tables.visit_dimension, records)



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.

from i2fhirb2.i2b2model.metadata.commondimension import CommonDimension
from i2fhirb2.sqlsupport.dynobject import DynObject, DynElements


class ModifierDimension(CommonDimension):
    _t = DynElements(CommonDimension)

    @DynObject.entry(_t)
    def modifier_path(self) -> str:
        return self.path()

    @DynObject.entry(_t)
    def modifier_cd(self) -> str:
        return self.cd()

    @DynObject.entry(_t)
    def name_char(self) -> str:
        return self.name_char_()

    @DynObject.entry(_t)
    def modifier_blob(self) -> str:
        return self.blob()

    def __lt__(self, other):
        return self.modifier_path < other.modifier_path

    def __eq__(self, other):
        return self.modifier_path == other.modifier_path


class ModifierDimensionRoot(ModifierDimension):
    _t = DynElements(ModifierDimension)

    def __init__(self, base: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self._base = base

    @DynObject.entry(_t)
    def modifier_path(self) -> str:
        return '\\' + self._base + '\\'

    @DynObject.entry(_t)
    def modifier_cd(self) -> str:
        return self._base

    @DynObject.entry(_t)
    def name_char(self) -> str:
        return self._base + " root"

    @DynObject.entry(_t)
    def modifier_blob(self) -> str:
        return ''



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from abc import ABCMeta, abstractmethod
from typing import Optional, List, Dict, Tuple, Set, NamedTuple

from i2fhirb2.fhir.fhirmetadatavocabulary import FHIRMetaDataVocabulary, FMVGraphNode, FMVGraphEdge
from rdflib import Graph, URIRef

from i2fhirb2.fhir.fhirspecific import composite_uri, concept_path, modifier_path
from i2fhirb2.i2b2model.metadata.i2b2ontology import OntologyEntry


class FMVGraphNodeWithMultiplicity(NamedTuple):
    graph_node: "FMVGraphNode"
    is_multiple: bool


class FHIRMetadata(metaclass=ABCMeta):
    """
    The W5 OntologyEntry in i2b2 space
    """

    def __init__(self, g: Graph, name_base: Optional[str]=None, modifier_base: Optional[str]=None) -> None:
        """
        Create a FHIR i2b2 generator
        :param g: Graph containing fhir.ttl information
        :param name_base: Base for concept_dimension / modifier dimension and non-modifier ontology paths
        :param modifier_base: base for modifier ontology paths (m_applied_path != '@')
        """
        if name_base is None:
            name_base = "\\FHIR\\"
        else:
            assert(name_base.endswith('\\'))
        if modifier_base is None:
            modifier_base = name_base[:-1] + "Mod\\"
        self._fhir_metadatavocabulary = FHIRMetaDataVocabulary()
        self._name_base = name_base
        self._modifier_base = modifier_base
        self.graph = g

    def _fhir_concept_expansion(self, concept_cd: URIRef, type_node: FMVGraphNode,
                                target: Dict[URIRef, FMVGraphNodeWithMultiplicity], nested: bool=False) -> None:
        """
        Expand all of the descendants of concept_cd until a leaf or a multiple_entries node is encountered
        :param concept_cd: Base resource predicate
        """
        target[concept_cd] = FMVGraphNodeWithMultiplicity(type_node, False)
        for edge in type_node.edges:
            extended_concept_cd = composite_uri(concept_cd, edge.predicate)
            if not edge.type_node.is_primitive:
                target[extended_concept_cd] = FMVGraphNodeWithMultiplicity(edge.type_node, edge.multiple_entries)
                if not edge.multiple_entries:
                    self._fhir_concept_expansion(extended_concept_cd, edge.type_node, target, nested=True)
            elif not nested:
                target[extended_concept_cd] = FMVGraphNodeWithMultiplicity(edge.type_node, edge.multiple_entries)

    def fhir_concepts(self, resource: Optional[URIRef]=None) -> Dict[URIRef, FMVGraphNodeWithMultiplicity]:
        """
        Return a list of resources in the FMV and the corresponding graph nodes.
        :param resource: Subject to restrict output to for debugging
        :return: A map from the concept URI to the corresponding Graph node
        """
        resources = [resource] if resource else self._fhir_metadatavocabulary.fhir_resource_concepts()
        rval = {}           # type: Dict[URIRef, FMVGraphNodeWithMultiplicity]
        for r in resources:
            self._fhir_concept_expansion(r, self._fhir_metadatavocabulary.resource_graph(r), rval)
        return rval

    def _add_modifier(self, modifier: FMVGraphEdge, target: Dict[URIRef, FMVGraphNode],
                      seen: Set[FMVGraphNode], depth: int=0) -> None:
        if modifier.type_node not in seen:
            if modifier.type_node.is_primitive:
                target[modifier.predicate] = modifier.type_node
            else:
                for edge in modifier.type_node.edges:
                    # seen.add(edge.type_node)
                    self._add_modifier(edge, target, seen, depth+1)

    def dimension_list(self, _: Optional[URIRef]=None) -> List[OntologyEntry]:
        """ Abstract class to return i2b2 dimension entries"""
        return []



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.


# TODO: Convert metadata_xml to json and update the clients
# TODO: Formal schema vs. template
from datetime import datetime
from typing import Optional, List, Tuple

from rdflib import URIRef

from i2fhirb2.fhir.fhirspecific import FHIR

xml_template = """<?xml version="1.0"?>
<ValueMetadata>
    <Version>3.02</Version>
    <CreationDateTime>{creation_date_time}</CreationDateTime>
    <TestID>{c_basecode}</TestID>
    <TestName>{c_name}</TestName>
    <DataType>{datatype}</DataType>
    <Flagstouse/>
    <Oktousevalues>Y</Oktousevalues>
    <EnumValues>{enum_list}</EnumValues>
    <UnitValues/>
</ValueMetadata>"""

enum_entry_template = '<Val description="{}">{}</Val>'

# <CreationDateTime>03/31/2011 11:15:00</CreationDateTime>
# TODO: find the correct mapping for these
# TODO: What to do with link?
time_type = "String"
date_type = "String"
datetime_type = "String"
instant_type = "String"
id_type = "String"
bool_type = "Bool"
value_type = "String"

# TODO: implement enum
fhir_type_map = {
    FHIR.decimal: "Float",
    FHIR.nodeRole: None,
    FHIR.time: time_type,
    FHIR.positiveInt: "PosInteger",
    FHIR.dateTime: datetime_type,
    FHIR.boolean: bool_type,
    # TODO: merge these once we add value set access
    # FHIR.code: "Enum",
    FHIR.realcode: "Enum",
    FHIR.code: "String",
    FHIR.base64Binary: None,
    FHIR.markdown: "largestring",
    FHIR.date: date_type,
    FHIR.treeRoot: None,
    FHIR.link: None,
    FHIR.integer: "Integer",
    FHIR.xhtml: "largestring",
    FHIR.instant: instant_type,
    FHIR.id: id_type,
    FHIR.string: "String",
    FHIR.uri: "String",
    FHIR.value: value_type,
    FHIR.unsignedInt: "Integer",
    FHIR.index: None,
    FHIR.oid: "String",
    FHIR.Reference: "String"
}

enum_bool_values = [("True value", "True"), ("False value", "False")]


def metadata_xml(typ: URIRef, c_basecode: str, c_name: str, creation_date: datetime,
                 pos_values: Optional[List[Tuple[str, str]]] = None) -> \
        Optional[str]:
    """
    Return the metadata xml for the given data type
    :param typ: FHIR primitive type
    :param c_basecode: i2b2 code
    :param c_name: i2b2 name
    :param creation_date: date to put into the metaxml field
    :param pos_values: list of possible values if code type
    :return: XML representation of element or None if the type is not to be realized
    """
    datatype = fhir_type_map.get(typ)
    if datatype is None:
        return None

    creation_date_time = creation_date
    if datatype == "Bool":
        pos_values = enum_bool_values
        datatype = "Enum"
    if datatype == "Enum" and pos_values is not None:
        enum_list = '\n        ' + \
                    '\n        '.join(enum_entry_template.format(desc, val) for desc, val in pos_values) + "\n    "
    else:
        enum_list = ""
    return xml_template.format(**locals())



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from typing import Optional, Dict, Tuple

from fhirtordf.rdfsupport.fhirgraphutils import link
from fhirtordf.rdfsupport.namespaces import FHIR
from rdflib import URIRef, Graph


class FHIR_Resource_type:

    def fact_key_for(self, g: Graph, subj: URIRef) -> Tuple[Optional[URIRef], Optional[URIRef], Optional[URIRef]]:
        return None, None, None


class FHIR_Infrastructure_type(FHIR_Resource_type):
    pass


class FHIR_Observation_Fact_type(FHIR_Resource_type):
    def __init__(self, patient_ref: URIRef, encounter_ref: Optional[URIRef], provider_ref: Optional[URIRef]) -> None:
        self.patient_ref = patient_ref
        self.encounter_ref = encounter_ref
        self.provider_ref = provider_ref

    def fact_key_for(self, g: Graph, subj: URIRef) -> Tuple[Optional[URIRef], Optional[URIRef], Optional[URIRef]]:
        pat_uri, pat_type = link(g, subj, self.patient_ref)
        # TODO: what should be done about unknown resources (Type RESOURCE)
        if pat_type and pat_type not in (FHIR.Patient, FHIR.Resource):
            return None, None, None
        if self.encounter_ref:
            encounter_uri, encounter_type = link(g, subj, self.encounter_ref)
            # TODO: Figure out what types of things count here
        else:
            encounter_uri = None
        if self.provider_ref:
            provider_uri, provider_type = link(g, subj, self.provider_ref)
            # TODO: Figure out what counts as a provider
            # if provider_type not in {FHIR.Organization}:
            #   provider_uri = None
        else:
            provider_uri = None
        return pat_uri, encounter_uri, provider_uri


class FHIR_Visit_Dimension_type(FHIR_Resource_type):
    def __init__(self, patient_ref: str) -> None:
        self.patient_ref = patient_ref


class FHIR_Provider_Dimension_type(FHIR_Resource_type):
    pass


class FHIR_Patient_Dimension_type(FHIR_Resource_type):

    def fact_key_for(self, g: Graph, subj: URIRef) -> Tuple[Optional[URIRef], Optional[URIRef], Optional[URIRef]]:
        return subj, None, link(g, subj, FHIR.managingOrganization)[0]


class FHIR_Bundle_type(FHIR_Resource_type):
    pass


FHIR_RESOURCE_MAP = {           # type: Dict[URIRef, Optional[FHIR_Resource_type]]
    FHIR.Account: FHIR_Observation_Fact_type(FHIR.Account.subject, None, FHIR.Account.owner),
    FHIR.ActivityDefinition: FHIR_Infrastructure_type(),
    FHIR.AdverseEvent: FHIR_Observation_Fact_type(FHIR.AdverseEvent.subject, None, None),
    FHIR.AllergyIntolerance: FHIR_Observation_Fact_type(FHIR.AllergyIntolerance.patient,
                                                        None,
                                                        FHIR.AllergyIntolerance.asserter),
    FHIR.Appointment: None,
    FHIR.AppointmentResponse: None,
    FHIR.AuditEvent: None,
    FHIR.Basic: None,
    FHIR.Binary: None,
    FHIR.BodyStructure: None,
    FHIR.Bundle: FHIR_Bundle_type,
    FHIR.CapabilityStatement: None,
    FHIR.CarePlan: FHIR_Observation_Fact_type(FHIR.CarePlan.subject, FHIR.CarePlan.context, None),
    FHIR.CareTeam: None,
    FHIR.ChargeItem: None,
    FHIR.Claim: None,
    FHIR.ClaimResponse: None,
    FHIR.ClinicalImpression: None,
    FHIR.CodeSystem: None,
    FHIR.Communication: None,
    FHIR.CommunicationRequest: None,
    FHIR.CompartmentDefinition: None,
    FHIR.Composition: None,
    FHIR.ConceptMap: None,
    FHIR.Condition: FHIR_Observation_Fact_type(FHIR.Condition.subject, FHIR.Condition.context, FHIR.Condition.asserter),
    FHIR.Consent: None,
    FHIR.Contract: None,
    FHIR.Coverage: None,
    FHIR.DetectedIssue: None,
    FHIR.Device: None,
    FHIR.DeviceComponent: None,
    FHIR.DeviceMetric: None,
    FHIR.DeviceRequest: None,
    FHIR.DeviceUseStatement: None,
    FHIR.DiagnosticReport: FHIR_Observation_Fact_type(FHIR.DiagnosticReport.subject, FHIR.DiagnosticReport.context,
                                                      FHIR.DiagnosticReport.performer),
    FHIR.DocumentManifest: None,
    FHIR.DocumentReference: None,
    FHIR.EligibilityRequest: None,
    FHIR.EligibilityResponse: None,
    FHIR.Encounter: FHIR_Visit_Dimension_type(FHIR.Encounter.subject),
    FHIR.Endpoint: None,
    FHIR.EnrollmentRequest: None,
    FHIR.EnrollmentResponse: None,
    FHIR.EpisodeOfCare: None,
    FHIR.EventDefinition: None,
    FHIR.ExpansionProfile: None,
    FHIR.ExplanationOfBenefit: None,
    FHIR.FamilyMemberHistory: None,
    FHIR.Flag: None,
    FHIR.Goal: None,
    FHIR.GraphDefinition: None,
    FHIR.Group: None,
    FHIR.GuidanceResponse: None,
    FHIR.HealthcareService: None,
    FHIR.ImagingManifest: None,
    FHIR.ImagingStudy: None,
    FHIR.Immunization: None,
    FHIR.ImmunizationRecommendation: None,
    FHIR.ImplementationGuide: None,
    FHIR.Library: None,
    FHIR.Linkage: None,
    FHIR.List: None,
    FHIR.Location: None,
    FHIR.Measure: None,
    FHIR.MeasureReport: FHIR_Infrastructure_type(),
    FHIR.Media: FHIR_Infrastructure_type(),
    FHIR.Medication: FHIR_Infrastructure_type(),
    FHIR.MedicationAdministration: FHIR_Observation_Fact_type(FHIR.MedicationAdministration.subject,
                                                              FHIR.MedicationAdministration.context,
                                                              FHIR.MedicationAdministration.performer),
    FHIR.MedicationDispense: FHIR_Observation_Fact_type(FHIR.MedicationDispense.subject,
                                                        FHIR.MedicationDispense.context,
                                                        FHIR.MedicationDispense.performer),
    FHIR.MedicationRequest: FHIR_Observation_Fact_type(FHIR.MedicationRequest.subject,
                                                       FHIR.MedicationRequest.context,
                                                       FHIR.MedicationRequest.requester),
    FHIR.MedicationStatement: FHIR_Observation_Fact_type(FHIR.MedicationStatement.subject,
                                                         FHIR.MedicationStatement.context,
                                                         None),
    FHIR.MessageDefinition: FHIR_Infrastructure_type(),
    FHIR.MessageHeader: FHIR_Infrastructure_type(),
    FHIR.NamingSystem: FHIR_Infrastructure_type(),
    FHIR.NutritionOrder: FHIR_Observation_Fact_type(FHIR.NutritionOrder.patient,
                                                    FHIR.NutritionOrder.encounter,
                                                    FHIR.NutritionOrder.orderer),
    FHIR.Observation: FHIR_Observation_Fact_type(FHIR.Observation.subject,
                                                 FHIR.Observation.context,
                                                 FHIR.Observation.performer),
    FHIR.OperationDefinition: FHIR_Infrastructure_type(),
    FHIR.OperationOutcome: FHIR_Infrastructure_type(),
    FHIR.Organization: FHIR_Provider_Dimension_type(),
    FHIR.Patient: FHIR_Patient_Dimension_type(),
    FHIR.PaymentNotice: None,
    FHIR.PaymentReconciliation: None,
    FHIR.Person: None,
    FHIR.PlanDefinition: None,
    FHIR.Practitioner: None,
    FHIR.PractitionerRole: None,
    FHIR.Procedure: None,
    FHIR.ProcedureRequest: None,
    FHIR.ProcessRequest: None,
    FHIR.ProcessResponse: None,
    FHIR.Provenance: None,
    FHIR.Questionnaire: None,
    FHIR.QuestionnaireResponse: None,
    FHIR.RelatedPerson: None,
    FHIR.RequestGroup: None,
    FHIR.ResearchStudy: None,
    FHIR.ResearchSubject: None,
    FHIR.RiskAssessment: None,
    FHIR.Schedule: None,
    FHIR.SearchParameter: None,
    FHIR.Sequence: None,
    FHIR.ServiceDefinition: None,
    FHIR.Slot: None,
    FHIR.Specimen: None,
    FHIR.StructureDefinition: None,
    FHIR.StructureMap: None,
    FHIR.Subscription: None,
    FHIR.Substance: None,
    FHIR.SupplyDelivery: None,
    FHIR.SupplyRequest: None,
    FHIR.Task: None,
    FHIR.TestReport: FHIR_Infrastructure_type(),
    FHIR.TestScript: FHIR_Infrastructure_type(),
    FHIR.ValueSet: FHIR_Infrastructure_type(),
    FHIR.VisionPrescription: FHIR_Observation_Fact_type(FHIR.VisionPrescription.patient,
                                                        FHIR.VisionPrescription.encounter,
                                                        FHIR.VisionPrescription.prescriber)
}



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from datetime import datetime
from operator import or_
from typing import Optional, List, Tuple, Callable, Dict

import sys
from sqlalchemy import Table, and_, update, delete, select
from sqlalchemy.engine import Connection

from i2fhirb2.i2b2model.shared.listchunker import ListChunker
from i2fhirb2.sqlsupport.dynobject import DynObject, DynElements, DynamicPropType


class I2B2_Core(DynObject):
    _t = DynElements()
    _check_dups = False

    def __init__(self,
                 update_date: Optional[DynamicPropType] = None,
                 download_date: Optional[DynamicPropType] = None,
                 sourcesystem_cd: Optional[DynamicPropType] = None,
                 import_date: Optional[DynamicPropType] = None):
        self._update_date = update_date
        self._download_date = download_date
        self._sourcesystem_cd = sourcesystem_cd
        self._import_date = import_date

    @DynObject.entry(_t)
    def update_date(self) -> datetime:
        return self._resolve(self._update_date) if self._update_date is not None else datetime.now()

    @DynObject.entry(_t)
    def download_date(self) -> datetime:
        return self._resolve(self._download_date) if self._download_date is not None else self.update_date

    @DynObject.entry(_t)
    def import_date(self) -> datetime:
        return self._resolve(self._import_date) if self._import_date is not None else self.update_date

    @DynObject.entry(_t)
    def sourcesystem_cd(self) -> str:
        return self._resolve(self._sourcesystem_cd) if self._sourcesystem_cd is not None else "Unspecified"


class I2B2_Core_With_Upload_Id(I2B2_Core):
    _t = DynElements(I2B2_Core)

    no_update_fields = ["update_date", "download_date", "import_date", "sourcesystem_cd", "upload_id"]
    key_fields = None

    def __init__(self,
                 upload_id: Optional[DynamicPropType] = None,
                 **kwargs):
        super().__init__(**kwargs)
        self._upload_id = upload_id

    DynObject._after_root(_t)

    @DynObject.entry(_t)
    def upload_id(self) -> Optional[int]:
        return self._resolve(self._upload_id)

    @staticmethod
    def _delete_upload_id(conn: Connection, table: Table, upload_id: int) -> int:
        """
        Remove all table records with the supplied upload_id
        :param conn: sql connection
        :param table: table to modify
        :param upload_id: target upload_id
        :return: number of records removed
        """
        return conn.execute(delete(table).where(table.c.upload_id == upload_id)).rowcount if upload_id else 0

    @staticmethod
    def _delete_sourcesystem_cd(conn: Connection, table: Table, sourcesystem_cd: str) -> int:
        """
        Remove all table records with the supplied upload_id
        :param conn: sql connection
        :param table: table to modify
        :param sourcesystem_cd: target sourcesystem code
        :return: number of records removed
        """
        return conn.execute(delete(table).where(table.c.sourcesystem_cd == sourcesystem_cd)).rowcount \
            if sourcesystem_cd else 0

    @staticmethod
    def _nested_fcn(f: Callable, filters: List):
        """
        Distribute binary function f across list L
        :param f: Binary function
        :param filters: function arguments
        :return: chain of binary filters
        """
        return None if len(filters) == 0 \
            else filters[0] if len(filters) == 1 \
            else f(filters[0], I2B2_Core_With_Upload_Id._nested_fcn(f, filters[1:]))

    @classmethod
    def _check_for_dups(cls, records: List["I2B2_Core_With_Upload_Id"]) -> \
            Dict[Tuple, List["I2B2_Core_With_Upload_Id"]]:
        key_map = dict()    # type: Dict[Tuple, I2B2_Core_With_Upload_Id]
        dups = dict()       # type: Dict[Tuple, List[I2B2_Core_With_Upload_Id]]
        for record in records:
            key = tuple(record.get(k) for k in cls.key_fields)
            if key in key_map:
                dups.setdefault(key, [key_map[key]]).append(record)
            else:
                key_map[key] = record
        return dups

    @classmethod
    def _add_or_update_records(cls, conn: Connection, table: Table,
                               records: List["I2B2_Core_With_Upload_Id"]) -> Tuple[int, int]:
        """
        Add or update the supplied table as needed to reflect the contents of records
        :param table: i2b2 sql connection
        :param records: records to apply
        :return: number of records added / modified
        """
        num_updates = 0
        num_inserts = 0
        inserts = []
        # Iterate over the records doing updates
        # Note: This is slow as molasses - definitely not optimal for batch work, but hopefully we'll be dealing with
        #    thousands to tens of thousands of records.  May want to move to ORM model if this gets to be an issue
        for record in records:
            keys = [(table.c[k] == getattr(record, k)) for k in cls.key_fields]
            key_filter = I2B2_Core_With_Upload_Id._nested_fcn(and_, keys)
            rec_exists = conn.execute(select([table.c.upload_id]).where(key_filter)).rowcount
            if rec_exists:
                known_values = {k: v for k, v in record._freeze().items()
                                if v is not None and k not in cls.no_update_fields and
                                k not in cls.key_fields}
                vals = [table.c[k] != v for k, v in known_values.items()]
                val_filter = I2B2_Core_With_Upload_Id._nested_fcn(or_, vals)
                known_values['update_date'] = record.update_date
                upd = update(table).where(and_(key_filter, val_filter)).values(known_values)
                num_updates += conn.execute(upd).rowcount
            else:
                inserts.append(record._freeze())
        if inserts:
            if cls._check_dups:
                dups = cls._check_for_dups(inserts)
                nprints = 0
                # TODO: Figure out why duplicates are occuring -- they are very specific
                if dups:
                    print("{} duplicate records encountered".format(len(dups)))
                    for k, vals in dups.items():
                        if len(vals) == 2 and vals[0] == vals[1]:
                            inserts.remove(vals[1])
                        else:
                            if nprints < 20:
                                print("Key: {} has a non-identical dup".format(k))
                            elif nprints == 20:
                                print(".... more ...")
                            nprints += 1
                            for v in vals[1:]:
                                inserts.remove(v)
            # TODO: refactor this to load on a per-resource basis.  Temporary fix
            for insert in ListChunker(inserts, 500):
                num_inserts += conn.execute(table.insert(), insert).rowcount
        return num_inserts, num_updates



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
import inspect
from collections import OrderedDict
from typing import List, Callable, Union

from i2fhirb2.sqlsupport.orderedset import OrderedSet

EvalParam = Union[Callable[[object], object], Callable[[], object], object]


class ColumnsBase:
    """
    Base class for a dynamically evaluated columns list
    """
    _columns = []                 # type: OrderedSet

    @classmethod
    def _bind_keys(cls, keys: List[str]):
        for k in keys:
            if k not in cls._columns:
                cls._columns.add(k)

    def _freeze(self) -> OrderedDict:
        """
        Evaluate all of the column values and return the result
        :return: column/value tuples
        """
        return OrderedDict(**{k: getattr(self, k, None) for k in super().__getattribute__("_columns")})

    def _eval(self, m: EvalParam) -> object:
        """
        Evaluate m returning the method / function invocation or value.  Kind of like a static method
        :param m: object to evaluate
        :return: return
        """
        if inspect.ismethod(m) or inspect.isroutine(m):
            return m()
        elif inspect.isfunction(m):
            return m(self) if len(inspect.signature(m)) > 0 else m()
        else:
            return m

    def __getattribute__(self, item):
        if item.startswith("_"):
            return super().__getattribute__(item)
        cols = super().__getattribute__("_columns")
        if item in cols:
            return self._eval(super().__getattribute__(item))
        return super().__getattribute__(item)



# Copyright (c) 2017, Mayo Clinic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice, this
#     list of conditions and the following disclaimer.
#
#     Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#
#     Neither the name of the Mayo Clinic nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
from collections import OrderedDict
from typing import Any, Dict, Tuple, Callable, Union, List, Type

DynamicPropType = Union[Type[classmethod], Type[staticmethod], Callable[[object], object], object]


class DynElements:
    """
    A map from an element name to either a function that returns an object or an object itself.
    :param parent: parent DynObject
    """
    _delimiter = '\t'

    def __init__(self, parent: Type["DynObject"]=None) -> None:
        """
        Construct a set of dynamic elements
        :param parent:
        """
        self.elements = OrderedDict()               # type: OrderedDict[str, Tuple[bool, DynamicPropType]]
        self.override_elements = OrderedDict()      # type: OrderedDict[str, Tuple[bool, DynamicPropType]]
        self.parent_dynelements = parent._t if parent else None

    def add(self, c: DynamicPropType) -> DynamicPropType:
        """
        Add an method to the set
        :param c: method to add
        :return: added method
        """
        fname = c.__func__.__name__ if isinstance(c, (classmethod, staticmethod)) else c.__name__
        self.elements[fname] = (True, c)
        return c

    def add_func(self, f: Callable[[], Any]) -> Callable[[], Any]:
        """
        Add a function to the set
        :param f: function to add
        :return: added function
        """
        self.elements[f.__name__] = (False, f)
        return f

    def add_const(self, name: str, c: Any) -> Any:
        """
        Add a constant to the set
        :param name:
        :param c:
        :return:
        """
        self.elements[name] = (False, c)
        return c

    def add_override_const(self, name: str, c: Any) -> Any:
        self.override_elements[name] = (False, c)
        return c

    def add_parents(self):
        """ Indicator to insert the parent dynelements here instead of at the end """
        self.elements[""] = (False, None)

    def clear_overrides(self):
        self.override_elements.clear()

    def clear_override(self, item):
        if item in self.override_elements:
            del(self.override_elements[item])

    def has_entry(self, name: str) -> bool:
        """
        Determine whether name is in the list
        :param name: name to test
        :return:
        """
        return name in self.elements or (self.parent_dynelements and self.parent_dynelements.has_entry(name))

    @staticmethod
    def resolve_entry(context: object, v: DynamicPropType, has_context: bool) -> Any:
        if isinstance(v, classmethod):
            return v.__func__(context)
        elif isinstance(v, staticmethod):
            return v.__func__()
        elif callable(v):
            return v(context) if has_context else v()
        return v

    def get_entry(self, context: object, name: str):
        if name in self.elements or name in self.override_elements:
            has_context, v = self.override_elements[name] if name in self.override_elements else self.elements[name]
            return self.resolve_entry(context, v, has_context)
        elif self.parent_dynelements:
            return self.parent_dynelements.get_entry(context, name)
        else:
            raise AttributeError("{} Not Found".format(name))

    def keys(self) -> List[str]:
        parent_keys = [k for k in self.parent_dynelements.keys()] if self.parent_dynelements else []
        rval = []
        for k in self.elements.keys():
            if k == "":
                rval += parent_keys
            elif k not in parent_keys:
                rval.append(k)
        if "" not in self.elements.keys():
            rval += parent_keys
        return rval

    def freeze(self, context: object) -> Dict[str, Any]:
        """
        Resolve all of the functions and return a set of values
        :param context: Reification context
        :return: ordered dictionary of reified results
        """
        return OrderedDict(**{k: self.get_entry(context, k) for k in self.keys()})

    @property
    def header(self):
        return self._delimiter.join(self.keys())


class DynObjectMetaClass(type):
    _t = None                   # type: DynElements

    def __setattr__(self, key, value):
        """
        Set a dynamic value.  This is always local - it will not replace the parent attribute
        """
        if self._t.has_entry(key):
            self._t.add_override_const(key, value)
        else:
            super().__setattr__(key, value)

    def __delitem__(self, item):
        """
        Remove an attribute from elements it it exists
        :param item:
        :return:
        """
        if self._t.has_entry(item):
            self._t.clear_override(item)
        else:
            super().__delattr__(item)


class DynObject(metaclass=DynObjectMetaClass):

    """
    A "Dynamic Object' - a class whose
    """
    _t = None           # type: DynElements
    _null_text = ''

    @staticmethod
    def entry(dynelements: DynElements) -> Callable[[object], Any]:
        def f(c: DynamicPropType):
            return dynelements.add(c)
        return f

    @classmethod
    def _clear(cls):
        cls._t.clear_overrides()

    def __setattr__(self, key, value):
        """
        Set a dynamic value.  This is always local - it will not replace the parent attribute
        """
        elements = super().__getattribute__('_t')
        if elements.has_entry(key):
            raise ValueError("Attributes must be changed on class level")
        else:
            super().__setattr__(key, value)

    def __getattribute__(self, item: str):
        """
        Translate distinguished elements into attributes
        :param item: name of item to retrieve
        :return: value for item
        """
        elements = super().__getattribute__('_t')
        if elements.has_entry(item):
            return elements.get_entry(self, item)
        else:
            return super().__getattribute__(item)

    def _freeze(self) -> Dict[str, Any]:
        """
        Return the distinguished list, removing all functional dependencies
        :return:
        """
        return self._t.freeze(self)

    @classmethod
    def _after_root(cls, dynelements: DynElements) -> None:
        dynelements.add_parents()

    @classmethod
    def _header(cls) -> str:
        return cls._t.header

    def _resolve(self, f: DynamicPropType) -> Any:
        return self._t.resolve_entry(self, f, False)

    @classmethod
    def _escape(cls, txt: str) -> str:
        return txt.replace(cls._t._delimiter, '\\' + cls._t._delimiter)

    def __str__(self):
        return "DynObject({})".format(', '.join(["{}:'{}'".format(k, getattr(self, k)) for k in self._t.keys()]))

    def __repr__(self):
        return self._t._delimiter.join(str(e) if e is not None else self._null_text for e in self._freeze().values())

    def __lt__(self, other):
        return repr(self) < repr(other)

    def __eq__(self, other):
        return repr(self) == repr(other)
